import { isBuiltin } from 'node:module';
import { IGNORED_DEPENDENCIES, IGNORED_GLOBAL_BINARIES, IGNORED_RUNTIME_DEPENDENCIES, IGNORE_DEFINITELY_TYPED, ROOT_WORKSPACE_NAME, } from './constants.js';
import { getDependencyMetaData } from './manifest/index.js';
import { getDefinitelyTypedFor, getPackageFromDefinitelyTyped, isDefinitelyTyped } from './util/modules.js';
import { hasMatch, hasMatchInSet, toRegexOrString } from './util/regex.js';
export class DependencyDeputy {
    isProduction;
    isStrict;
    _manifests = new Map();
    referencedDependencies;
    referencedBinaries;
    hostDependencies;
    installedBinaries;
    hasTypesIncluded;
    constructor({ isProduction, isStrict }) {
        this.isProduction = isProduction;
        this.isStrict = isStrict;
        this.referencedDependencies = new Map();
        this.referencedBinaries = new Map();
        this.hostDependencies = new Map();
        this.installedBinaries = new Map();
        this.hasTypesIncluded = new Map();
    }
    addWorkspace({ name, cwd, dir, manifestPath, manifest, ignoreDependencies, ignoreBinaries, }) {
        const dependencies = Object.keys(manifest.dependencies ?? {});
        const peerDependencies = Object.keys(manifest.peerDependencies ?? {});
        const optionalDependencies = Object.keys(manifest.optionalDependencies ?? {});
        const optionalPeerDependencies = manifest.peerDependenciesMeta
            ? peerDependencies.filter(peerDependency => manifest.peerDependenciesMeta &&
                peerDependency in manifest.peerDependenciesMeta &&
                manifest.peerDependenciesMeta[peerDependency].optional)
            : [];
        const devDependencies = Object.keys(manifest.devDependencies ?? {});
        const allDependencies = [...dependencies, ...devDependencies, ...peerDependencies, ...optionalDependencies];
        const packageNames = [
            ...dependencies,
            ...(this.isStrict ? peerDependencies : []),
            ...(this.isProduction ? [] : devDependencies),
        ];
        const { hostDependencies, installedBinaries, hasTypesIncluded } = getDependencyMetaData({
            packageNames,
            dir,
            cwd,
        });
        this.setHostDependencies(name, hostDependencies);
        this.setInstalledBinaries(name, installedBinaries);
        this.setHasTypesIncluded(name, hasTypesIncluded);
        this._manifests.set(name, {
            workspaceDir: dir,
            manifestPath,
            ignoreDependencies: ignoreDependencies.map(toRegexOrString),
            ignoreBinaries: ignoreBinaries.map(toRegexOrString),
            dependencies,
            devDependencies,
            peerDependencies: new Set(peerDependencies),
            optionalPeerDependencies,
            allDependencies: new Set(allDependencies),
        });
    }
    getWorkspaceManifest(workspaceName) {
        return this._manifests.get(workspaceName);
    }
    getProductionDependencies(workspaceName) {
        const manifest = this._manifests.get(workspaceName);
        if (!manifest)
            return [];
        if (this.isStrict)
            return [...manifest.dependencies, ...manifest.peerDependencies];
        return manifest.dependencies;
    }
    getDevDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.devDependencies ?? [];
    }
    getDependencies(workspaceName) {
        const manifest = this._manifests.get(workspaceName);
        if (!manifest)
            return new Set();
        return new Set([...manifest.dependencies, ...manifest.devDependencies]);
    }
    setInstalledBinaries(workspaceName, installedBinaries) {
        this.installedBinaries.set(workspaceName, installedBinaries);
    }
    getInstalledBinaries(workspaceName) {
        return this.installedBinaries.get(workspaceName);
    }
    setHasTypesIncluded(workspaceName, hasTypesIncluded) {
        this.hasTypesIncluded.set(workspaceName, hasTypesIncluded);
    }
    getHasTypesIncluded(workspaceName) {
        return this.installedBinaries.get(workspaceName);
    }
    addReferencedDependency(workspaceName, packageName) {
        if (!this.referencedDependencies.has(workspaceName)) {
            this.referencedDependencies.set(workspaceName, new Set());
        }
        this.referencedDependencies.get(workspaceName)?.add(packageName);
    }
    addReferencedBinary(workspaceName, binaryName) {
        if (!this.referencedBinaries.has(workspaceName)) {
            this.referencedBinaries.set(workspaceName, new Set());
        }
        this.referencedBinaries.get(workspaceName)?.add(binaryName);
    }
    setHostDependencies(workspaceName, hostDependencies) {
        this.hostDependencies.set(workspaceName, hostDependencies);
    }
    getHostDependenciesFor(workspaceName, dependency) {
        return this.hostDependencies.get(workspaceName)?.get(dependency) ?? [];
    }
    getPeerDependencies(workspaceName) {
        const manifest = this._manifests.get(workspaceName);
        if (!manifest)
            return new Set();
        return manifest.peerDependencies;
    }
    getOptionalPeerDependencies(workspaceName) {
        const manifest = this._manifests.get(workspaceName);
        if (!manifest)
            return [];
        return manifest.optionalPeerDependencies;
    }
    isIgnoredDependency(workspaceName, packageName) {
        const manifest = this.getWorkspaceManifest(workspaceName);
        if (manifest && hasMatch(manifest.ignoreDependencies, packageName))
            return true;
        if (workspaceName !== ROOT_WORKSPACE_NAME)
            return this.isIgnoredDependency(ROOT_WORKSPACE_NAME, packageName);
        return false;
    }
    isIgnoredBinary(workspaceName, binaryName) {
        const manifest = this.getWorkspaceManifest(workspaceName);
        if (manifest && hasMatch(manifest.ignoreBinaries, binaryName))
            return true;
        if (workspaceName !== ROOT_WORKSPACE_NAME)
            return this.isIgnoredBinary(ROOT_WORKSPACE_NAME, binaryName);
        return false;
    }
    maybeAddReferencedExternalDependency(workspace, packageName) {
        if (isBuiltin(packageName))
            return true;
        if (IGNORED_RUNTIME_DEPENDENCIES.has(packageName))
            return true;
        if (packageName === workspace.pkgName)
            return true;
        const workspaceNames = this.isStrict ? [workspace.name] : [workspace.name, ...[...workspace.ancestors].reverse()];
        const closestWorkspaceName = workspaceNames.find(name => this.isInDependencies(name, packageName));
        const typesPackageName = !isDefinitelyTyped(packageName) && getDefinitelyTypedFor(packageName);
        const closestWorkspaceNameForTypes = typesPackageName && workspaceNames.find(name => this.isInDependencies(name, typesPackageName));
        if (closestWorkspaceName || closestWorkspaceNameForTypes) {
            closestWorkspaceName && this.addReferencedDependency(closestWorkspaceName, packageName);
            closestWorkspaceNameForTypes && this.addReferencedDependency(closestWorkspaceNameForTypes, typesPackageName);
            return true;
        }
        this.addReferencedDependency(workspace.name, packageName);
        if (this.isIgnoredDependency(workspace.name, packageName))
            return true;
        return false;
    }
    maybeAddReferencedBinary(workspace, binaryName) {
        if (IGNORED_GLOBAL_BINARIES.has(binaryName))
            return true;
        this.addReferencedBinary(workspace.name, binaryName);
        const workspaceNames = this.isStrict ? [workspace.name] : [workspace.name, ...[...workspace.ancestors].reverse()];
        for (const name of workspaceNames) {
            const binaries = this.getInstalledBinaries(name);
            if (binaries?.has(binaryName)) {
                const dependencies = binaries.get(binaryName);
                if (dependencies?.size) {
                    for (const dependency of dependencies)
                        this.addReferencedDependency(name, dependency);
                    return true;
                }
            }
        }
        if (this.isIgnoredBinary(workspace.name, binaryName))
            return true;
        return false;
    }
    isInDependencies(workspaceName, packageName) {
        const manifest = this._manifests.get(workspaceName);
        if (!manifest)
            return false;
        if (this.isStrict)
            return this.getProductionDependencies(workspaceName).includes(packageName);
        return manifest.allDependencies.has(packageName);
    }
    settleDependencyIssues() {
        const dependencyIssues = [];
        const devDependencyIssues = [];
        const optionalPeerDependencyIssues = [];
        for (const [workspaceName, { manifestPath }] of this._manifests.entries()) {
            const referencedDependencies = this.referencedDependencies.get(workspaceName);
            const installedBinaries = this.getInstalledBinaries(workspaceName);
            const hasTypesIncluded = this.getHasTypesIncluded(workspaceName);
            const isNotIgnoredDependency = (packageName) => {
                if (this.isIgnoredDependency(workspaceName, packageName))
                    return false;
                return !IGNORED_DEPENDENCIES.has(packageName);
            };
            const isNotIgnoredBinary = (packageName) => {
                if (installedBinaries?.has(packageName)) {
                    const binaryNames = installedBinaries.get(packageName);
                    if (binaryNames) {
                        for (const b of binaryNames) {
                            if (this.isIgnoredBinary(workspaceName, b))
                                return false;
                            if (IGNORED_GLOBAL_BINARIES.has(b))
                                return false;
                        }
                    }
                }
                return true;
            };
            const peerDepRecs = {};
            const isReferencedDependency = (dependency, isPeerDep) => {
                if (referencedDependencies?.has(dependency))
                    return true;
                if (isPeerDep && peerDepRecs[dependency])
                    return false;
                const [scope, typedDependency] = dependency.split('/');
                if (scope === '@types') {
                    if (hasTypesIncluded?.has(typedDependency))
                        return false;
                    const typedPackageName = getPackageFromDefinitelyTyped(typedDependency);
                    if (IGNORE_DEFINITELY_TYPED.includes(typedPackageName))
                        return true;
                    const hostDependencies = [
                        ...this.getHostDependenciesFor(workspaceName, dependency),
                        ...this.getHostDependenciesFor(workspaceName, typedPackageName),
                    ];
                    if (hostDependencies.length)
                        return !!hostDependencies.find(host => isReferencedDependency(host.name, true));
                    if (!referencedDependencies)
                        return false;
                    return referencedDependencies.has(typedPackageName);
                }
                const hostDependencies = this.getHostDependenciesFor(workspaceName, dependency);
                for (const { name } of hostDependencies) {
                    if (!peerDepRecs[name])
                        peerDepRecs[name] = 1;
                    else
                        peerDepRecs[name]++;
                }
                return hostDependencies.some(hostDependency => (isPeerDep === false || !hostDependency.isPeerOptional) && isReferencedDependency(hostDependency.name, true));
            };
            const isNotReferencedDependency = (dependency) => !isReferencedDependency(dependency, false);
            const pd = this.getProductionDependencies(workspaceName);
            const dd = this.getDevDependencies(workspaceName);
            const od = this.getOptionalPeerDependencies(workspaceName);
            pd.filter(isNotIgnoredDependency)
                .filter(isNotIgnoredBinary)
                .filter(isNotReferencedDependency)
                .forEach(symbol => dependencyIssues.push({ type: 'dependencies', filePath: manifestPath, symbol }));
            dd.filter(isNotIgnoredDependency)
                .filter(isNotIgnoredBinary)
                .filter(isNotReferencedDependency)
                .forEach(symbol => devDependencyIssues.push({ type: 'devDependencies', filePath: manifestPath, symbol }));
            od.filter(isNotIgnoredDependency)
                .filter(isNotIgnoredBinary)
                .filter(p => isReferencedDependency(p))
                .forEach(symbol => optionalPeerDependencyIssues.push({ type: 'optionalPeerDependencies', filePath: manifestPath, symbol }));
        }
        return { dependencyIssues, devDependencyIssues, optionalPeerDependencyIssues };
    }
    getConfigurationHints() {
        const configurationHints = new Set();
        const allDependencies = new Set();
        const allPeerDependencies = new Set();
        const allReferencedDependencies = new Set();
        const allReferencedBinaries = new Set();
        const allInstalledBinaryNames = new Set();
        for (const [workspaceName, { ignoreDependencies, ignoreBinaries }] of this._manifests.entries()) {
            const dependencies = this.getDependencies(workspaceName);
            const peerDependencies = this.getPeerDependencies(workspaceName);
            const referencedDependencies = this.referencedDependencies.get(workspaceName) ?? new Set();
            const referencedBinaries = this.referencedBinaries.get(workspaceName) ?? new Set();
            const installedBinaries = this.getInstalledBinaries(workspaceName) ?? new Set();
            const installedBinaryNames = new Set(installedBinaries.keys());
            for (const id of dependencies)
                allDependencies.add(id);
            for (const id of peerDependencies)
                allPeerDependencies.add(id);
            for (const id of referencedDependencies)
                allReferencedDependencies.add(id);
            for (const id of referencedBinaries)
                allReferencedBinaries.add(id);
            for (const id of installedBinaryNames)
                allInstalledBinaryNames.add(id);
            if (workspaceName === ROOT_WORKSPACE_NAME)
                continue;
            for (const identifier of ignoreDependencies) {
                const isListed = hasMatchInSet(dependencies, identifier) && !hasMatchInSet(peerDependencies, identifier);
                const isReferenced = hasMatchInSet(referencedDependencies, identifier);
                if (typeof identifier === 'string' && isListed && isReferenced) {
                    configurationHints.add({ workspaceName, identifier, type: 'ignoreDependencies' });
                }
                else if (!this.isProduction && !isListed && !isReferenced) {
                    configurationHints.add({ workspaceName, identifier, type: 'ignoreDependencies' });
                }
            }
            for (const identifier of ignoreBinaries) {
                const isInstalled = hasMatchInSet(installedBinaryNames, identifier);
                const isReferenced = hasMatchInSet(referencedBinaries, identifier);
                if (typeof identifier === 'string' && isInstalled && isReferenced) {
                    configurationHints.add({ workspaceName, identifier, type: 'ignoreBinaries' });
                }
                else if (!this.isProduction && !isInstalled && !isReferenced) {
                    configurationHints.add({ workspaceName, identifier, type: 'ignoreBinaries' });
                }
            }
        }
        const manifest = this.getWorkspaceManifest(ROOT_WORKSPACE_NAME);
        if (manifest) {
            for (const identifier of manifest.ignoreDependencies) {
                const isListed = hasMatchInSet(allDependencies, identifier) && !hasMatchInSet(allPeerDependencies, identifier);
                const isReferenced = hasMatchInSet(allReferencedDependencies, identifier);
                if (typeof identifier === 'string' && !isDefinitelyTyped(identifier) && isListed && isReferenced) {
                    configurationHints.add({ workspaceName: ROOT_WORKSPACE_NAME, identifier, type: 'ignoreDependencies' });
                }
                else if (!this.isProduction && !isListed && !isReferenced) {
                    configurationHints.add({ workspaceName: ROOT_WORKSPACE_NAME, identifier, type: 'ignoreDependencies' });
                }
            }
            for (const identifier of manifest.ignoreBinaries) {
                const isInstalled = hasMatchInSet(allInstalledBinaryNames, identifier);
                const isReferenced = hasMatchInSet(allReferencedBinaries, identifier);
                if (typeof identifier === 'string' && isInstalled && isReferenced) {
                    configurationHints.add({ workspaceName: ROOT_WORKSPACE_NAME, identifier, type: 'ignoreBinaries' });
                }
                else if (!this.isProduction && !isInstalled && !isReferenced) {
                    configurationHints.add({ workspaceName: ROOT_WORKSPACE_NAME, identifier, type: 'ignoreBinaries' });
                }
            }
        }
        return configurationHints;
    }
}
